import os
import subprocess
from typing import List, Optional

import rich.progress as prog
from rich.progress import Progress

PROGBAR_COLUMNS = (
    prog.SpinnerColumn(),
    prog.TextColumn("[progress.description]{task.description}"),
    prog.BarColumn(),
    prog.TimeElapsedColumn(),
)


def get_zcat_command() -> str:
    """
    Checks if zcat or gzcat exist on the machine, returns whichever is functional!
    """
    try:
        subprocess.run("zcat --help", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError:
        try:
            subprocess.run("gzcat --help", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError:
            raise Exception(
                "Error when attempting to run zcat or gzcat command for k-mer counting."
            )
        else:
            return "gzcat"
    else:
        return "zcat"


def check_jellyfish():
    """
    Checks if Jellyfish is installed by running it's help command.

    Does nothing if Jellyfish is installed, raises an exception if Jellyfish is not found.
    """
    try:
        subprocess.check_output("jellyfish --help", shell=True)
    except subprocess.CalledProcessError:
        raise Exception("Jellyfish not found. Please install Jellyfish to count kmers.")


def try_command(command: str, err_msg: str):
    """
    Runs and command and if there's an error, raises an exception with the provided error message.
    """
    try:
        subprocess.run(command, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        raise subprocess.CalledProcessError(f"{err_msg}{e}")


def fastq_to_kmer_counts(
    input_paths: List[str],
    kmer_length: int,
    output_dir: Optional[str] = None,
    threads: int = 10,
    hash_size: int = 100_000_000,
) -> str:
    """_summary_

    Args:
        input_paths (List[str]): The input `fastq` or `fq` files to count.
        kmer_length (int): The length of the k-mers to count.
        output_dir (Optional[str], optional): The directory to put the counts file. Defaults to None (equivalent to current working directory).
        threads (int, optional): The number of threads to count with. Defaults to 10.
        hash_size (int, optional): The hash_size for Jellyfish counting. Defaults to 100_000_000.

    Raises:
        CalledProcessError: If a jellyfish command throws a runtime error.
        ValueError: If threads or hash_size or not a postiive integer
        NotADirectoryError: If output_dir is not a value directory

    Returns:
        str: The path of the output counts file
    """
    check_jellyfish()
    
    if type(kmer_length) is not int:
        raise TypeError("Jellyfish: kmer_length must be a positive integer.")
    elif kmer_length < 1:
        raise ValueError("Jellyfish: kmer_length must be a postive integer")
    
    if type(threads) is not int:
        raise TypeError("Jellyfish: Number of threads must be a positive integer.")
    elif threads < 1:
        raise ValueError("Jellyfish: Number of threads must be a postive integer")

    if type(hash_size) is not int:
        raise TypeError("Jellyfish: hash_size must be a positive integer.")
    elif threads < 1:
        raise ValueError("Jellyfish: hash_size must be a postive integer")
    
    # uses the accession number as the base name of the file
    base_path = f"{os.path.basename(input_paths[0].replace('_1', ''))}"

    if output_dir is not None:
        if not os.path.exists(output_dir):
            raise NotADirectoryError("Jellyfish: Counts file output directory does not exist.")

        base_path = f"{output_dir}/{base_path}"

    # Modify the file extension to .jf for the output
    jf_file = base_path.replace(".fastq", f"_{kmer_length}.jf").replace(".gz", "")

    # The base command for kmer counting
    count_command = (
        f"jellyfish count -m {kmer_length} -s {hash_size} -C -t {threads} -o {jf_file}"
    )

    # modifies the base command depending on if the files are zipped or not
    if input_paths[0].endswith(".fastq.gz"):
        zcat_command = get_zcat_command()
        count_command = (
            f"cat {' '.join(input_paths)} | {zcat_command} | {count_command} /dev/fd/0"
        )
    else:
        count_command = f"{count_command} {' '.join(input_paths)}"

    # Run count and dump jellyfish commands
    with Progress(*PROGBAR_COLUMNS) as progress:
        task = progress.add_task(f"Counting {kmer_length}-mers...", total=None)

        try_command(count_command, err_msg="Jellyfish: Error running 'jellyfish count' command: ")

        counts_file = jf_file.replace(".jf", ".counts")
        dump_command = f"jellyfish dump -c {jf_file} > {counts_file}"

        try_command(dump_command, err_msg="Jellyfish: Error running 'jellyfish dump' command: ")
        progress.update(task, total=1, advance=1)

    return counts_file


def try_fastq_to_kmer_counts(input_paths: List[str], **kwargs) -> str:
    """
    Attempts to run fastq_to_kmer_counts but catches the error if the keyward arguments throw an error.

    Args:
        input_paths (List[str]): The list of fastq input paths
    """
    try:
        return fastq_to_kmer_counts(input_paths, **kwargs)
    except Exception as e:
        if "unexpected keyword argument" in str(e):
            bad_key = str(e).split("'")[1]
            raise KeyError(f"Jellyfish: '{bad_key}' is not a valid argument for counting kmers.")
        elif "required" in str(e):
            bad_key = str(e).split("'")[1]
            raise KeyError(f"Jellyfish: '{bad_key}' argument is required for counting kmers.")
        else:
            raise e


import os
import sys
from typing import Optional
from importlib.metadata import version

import typer
from rich import print, panel

from pylagg import cgr, ena, jellyfish
import pylagg.config as conf

app = typer.Typer(add_completion=False)

def accession_to_cgr(accession: str, k: int, output_dir: str, threads: int, size : int):
    '''
    Takes an accession number and k-mer count and returns a CGR image. 

    NOTE: If accession refers to a project, will convert all runs in project to an image.
    '''
    run_accessions = ena.get_run_accessions([accession])

    for acc in run_accessions:
        files = ena.download_fastq(acc, output_dir) 
        counts_path = jellyfish.fastq_to_kmer_counts(files, k, output_dir, threads)

        with open(counts_path, 'r') as f:
            cgr.count_file_to_image_file(f, counts_path.replace(".counts", ".png"), size = size)


@app.command(name="cgr")
def cgr_command(
    input: str = typer.Option(
        None,
        "--input",
        "-i",
        help = "File name if using k-mer input you already have. Must be a .txt file for a single image.",
    ),
    accession_number: str = typer.Option(
        None,
        "--accession-number",
        "-a",
        help = 'Generate an image using an NCBI database accession number. If you would like to use more than one accession number, please list them in quotation marks, separated by commas, or put them in a .txt file with one accession number per line, and input the file name with this flag.',
    ),
    kmer: int = typer.Option(
        10,
        "--kmer",
        "-k",
        help = "Specify your desired k-mer length (Only used when generating from an accession number, if your input is already in k-mer form it will be detected)."
    ),
    output_dir: str = typer.Option(
        os.getcwd(),
        "--output-dir",
        "-o",
        help="Use this to specify an alternate save directory for your generated images. If nothing is specified, the default location is where the terminal is located.",
    ),
    size: Optional[int] = typer.Option(
        None,
        "--size",
        "-s",
        show_default=False,
        help = "Define an alternative side length for your image. Cannot exceed default value of 2^k.",
    ),
    thread_count: int = typer.Option(
        16,
        "--thread-count",
        "-t",
        help = "Amount of threads you wish to use. Threads are only used for k-mer counting when generating from an accession number.",
    ),
    config: str = typer.Option(
        None,
        "--config",
        "-c",
        help = "Use a config file to specify your options. Please include only an input (accession number(s) or kmer file(s)) and the config file's path. If any other options are also specified they will be ignored."
        ),
):
    """
    Generate your graph. Type "lagg cgr --help" to see all options for this command.
    """
    
    # INPUT ERROR CHECKING
    if input and accession_number:
        raise Exception("Please only include a file name OR an accession number(s).\nIf you need help, type 'lagg --help'.")
    # NOTE: melanie edited this to include config. pls change if you want a diff logic here @ollie!
    if not (input or accession_number or config):
        raise Exception("Please include an input, either an accession number(s), the name of a file containing k-mer input(s), or a config file path.\nIf you need help, type 'lagg --help'.")
    if size and (size > 2**kmer or size < 1):
        raise Exception("Your size is invalid. Please check and try again. Size cannot exceed 2^k, or be less than 1.\nIf you need help, type 'lagg --help'.")
    if kmer <= 0:
        raise Exception("Invalid k-mer count. Please use a k-mer count greater than 0. \nIf you need help, type 'lagg --help'.")
    if not(os.path.exists(output_dir)):
        raise Exception("The given output path does not exist. Please double check your path and try again. \nIf you need help, type 'lagg --help'.")
    
    #if no size is specified we need a default, and since it relies on other parameters it has to be done here.
    if not size:
        size = 2**kmer
    
    # Pre Process Accession number
    if config:
        conf.config_accession_to_cgr(config, output_dir)

    elif accession_number:
        
        if(accession_number.rfind(".txt") != -1):
            f = open(accession_number)
            for number in f:
                number = number.replace("\n", "")
                accession_to_cgr(number, kmer, output_dir, thread_count, size)
            f.close()
        else:
            #remove white spaces
            accession_number = accession_number.replace(" ", "")
            
            accession_list = accession_number.split(",")
            
            # Image generation with accession number:
            for number in accession_list:
                accession_to_cgr(number, kmer, output_dir, thread_count, size)
        
    elif input:

        input = input.replace(" ", "")
        inputlist = input.split(",")
        for file in inputlist:
            if not os.path.exists(file):
                raise Exception("There's an error with one of your input files " + file + ". Please double check your input and try again.")
        
        for file in inputlist:
            counts_file_path = file
            with open(counts_file_path) as f:
                
                # .txt case
                if(file.rfind(".txt") != -1):
                    file = file.replace(".txt", "")
                    
                    inputname = file[file.rfind("/") + 1:]
                    
                    cgr.count_file_to_image_file(f, output_dir + "/" + inputname + ".png", size=size)
                    print("Successfully created image called '" + inputname + ".png' at " + output_dir)
                else:
                    raise Exception("The input is not a supported file type. Supported types are '.txt'. Please convert to a supported file type and try again.\nIf you need help, type 'lagg --help'.")
        
    else:
        raise Exception("No valid flag. Please use either -a or -i or -c")
 

@app.command(name="ena")               
def ena_command(
    accession_number: str = typer.Option(
        None,
        "--accession-number",
        "-a",
        help = 'Download a fastq file using an NCBI database accession number. If you would like to use more than one accession number, please list them in quotation marks, separated by commas, or put them in a .txt file with one accession number per line, and input the file name with this flag.',
    ),
    kmer_length: Optional[int] = typer.Option(
        None,
        "--kmer",
        "-k",
        help = "If you would also like your download to be k-mer counted as well, provide a kmer length."
    ),
    output_dir: str = typer.Option(
        os.getcwd(),
        "--output-dir",
        "-o",
        help="Use this to specify a specific directory for your downloaded files (and k-mer count files, if counting). If nothing is specified, the default location is where the terminal is located.",
    ),
    thread_count: int = typer.Option(
        16,
        "--thread-count",
        "-t",
        help = "Amount of threads you wish to use. Threads are only used for k-mer counting.",
    ),
):
    """
    Only download a fastq file from ENA without generating a graph, and optionally k-mer count it too!
    """
    
    accession_number = accession_number.replace(" ", "")
        
    accession_list = accession_number.split(",")
        
    if(accession_number.rfind(".txt") != -1):
        raise Exception("detecting .txt of accession numbers (not implemented)")
    
    for number in accession_list:  
        files = ena.download_fastq(number, output_dir)
    
        if kmer_length is not None:
            print(files)
            jellyfish.fastq_to_kmer_counts(files, kmer_length, output_dir, thread_count)
            print("Successfully downloaded and created k-mer counts at " + output_dir)
        else:
            print("Successfully downloaded file(s) at " + output_dir)


@app.command(name="jellyfish")
def jellyfish_command(
    input_fastq: str = typer.Option(
        None,
        "--input-fastq",
        "-i",
        help = 'Convert a fastq file you already have downloaded. Input the file path and name using this flag.',
    ),
    kmer: int = typer.Option(
        10,
        "--kmer",
        "-k",
        help = "Specify your desired k-mer length for counting."
    ),
    output_dir: str = typer.Option(
        os.getcwd(),
        "--output-dir",
        "-o",
        help="Use this to specify an specific directory for your k-mer count files. If nothing is specified, the default location is where the terminal is located.",
    ),
    thread_count: int = typer.Option(
        16,
        "--thread-count",
        "-t",
        help = "Amount of threads you wish to use for your kmer counting.",
    ),
    ):
    """
    Generate a k-mer count from a fastq file already downloaded on your computer!
    """
    
    jellyfish.fastq_to_kmer_counts([input_fastq], kmer, output_dir, thread_count)


def version_callback(value: bool):
    if value:
        print(f"LAGG {version('pylagg')}")
        raise typer.Exit()


@app.callback()
def common(
    version: bool = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the current LAGG version and exit.",
    ),
):
    pass


def cli():
    try:
        app()
    except Exception as e:
        msg = str(e).strip('"')
        print(panel.Panel(msg, title="[red]Error", title_align="left", border_style="red"))
        sys.exit()



import os

from cutadapt.cli import main
from rich import progress as prog
from rich.progress import Progress

PROGBAR_COLUMNS = (
    prog.SpinnerColumn(),
    prog.TextColumn("[progress.description]{task.description}"),
    prog.BarColumn(),
    prog.TimeElapsedColumn()
)

def trim(input_file: list[str], config: dict):

    #Checks that the incoming list isn't empty
    if not input_file:
        raise Exception("No file names received from ENA download")
    
    #Checks that the file paths exist in the input_file list
    for files in input_file:
        if not os.path.isfile(files):
            raise Exception("Invalid files paths received from ENA download")
        

    command = []  # Initialize an empty list to hold the commands

    if len(input_file) == 2:
        command.append("--interleaved")
    
    
    arguments = config['cutadapt']
    
    # Iterate through each key and its associated list
    for key, value in arguments.items():
        if isinstance(value, list):
            for item in value:
                command.append(f"{key}")

                # Only append the item if it's not an empty string
                if item:
                    command.append(f"{item}")
        else:
            # Append the key
            command.append(f"{key}")

            # Check if value is not an empty list or empty string and append it
            if value not in ["", []]:
                command.append(f"{value}")
    
    if '-o' not in command:
        # Automatically generate the output file name by adding "_trimmed" before the extension
    
        name, ext = os.path.splitext(input_file[0])
        name, second_ext = os.path.splitext(name)
        name = name.rstrip("_1") #if there are 2 fastq files, will look at first file name, remove _1 from the end
        output_file= f"{name}_trimmed{second_ext}{ext}"
        command.append('-o')
        command.append(output_file)

    else:
        raise Exception("Can not create own output file name. Bulk runs will create only one file with that name. Please remove the -o flag in config file")

    for files in input_file:
        command.append(files)

    with Progress(*PROGBAR_COLUMNS) as progress:
        task = progress.add_task("Trimming with Cutadapt...", total=None)
        main(command)
        progress.update(task, total=1, advance=1)

    return output_file


import os
import toml
from typing import Any, Optional

from pylagg import ena, jellyfish, cgr
from pylagg.cutadapt import trim

class MissingConfigSection(KeyError): 
    pass

def load_toml_file(config_file: str) -> dict:
    """
    Reads a TOML file and returns a dictionary of the contents.
    """
    if not os.path.isfile(config_file):
        raise FileNotFoundError(
            f"No config file found at '{config_file}'. Please ensure you have given the correct path."
        )

    # read the TOML file
    try:
        config = toml.load(config_file)
    except Exception:
        raise toml.TomlDecodeError(
            "An error occurred while reading the TOML config file. Please ensure it is a valid TOML file."
        )

    return config

def get_config_key(config: dict, key: str) -> Any:
    """
    Checks if a config dictionary contains a key and it isn't empty.

    Args:
        config (dict): A dictionary of the config
        key (str): The key to search for

    Raises:
        KeyError: If the key is not found in the dictionary
        ValueError: If the key is found but is empty

    Returns:
        The value of the key in the config.
    """
    if key not in config:
        raise KeyError(f"Config error: {key} not found in config.")
    
    if config[key] is None or config[key] == {}:
        raise ValueError(f"Config error: {key} is empty.")
    
    return config[key]


def config_accession_to_cgr(config_file: str, output_dir: Optional[str] = None):
    """
    Takes a path to a config file and generates an image using the data.

    Args:
        config_file (str): The path to the config file
    """
    config = load_toml_file(config_file)

    download_args = get_config_key(config, "download")
    accession_arg = get_config_key(download_args, "accessions")
    accession_list = ena.get_run_accessions(accession_arg)
    
    for accession_number in accession_list:
        fastq_files = ena.download_fastq(accession_number, output_dir)

        jellyfish_args = get_config_key(config, "jellyfish")

        if output_dir is not None:
            jellyfish_args["output_dir"] = output_dir

        # Check if the 'cutadapt' section exists
        if "cutadapt" in config:
            trimmed_file = trim(fastq_files, config)
            
            counts_path = jellyfish.try_fastq_to_kmer_counts([trimmed_file], **jellyfish_args)

            # will activate if trim key is missing or equals "False"
            if not config.get("files", {}).get("trim", False):
                os.remove(trimmed_file)

        else:
            counts_path = jellyfish.try_fastq_to_kmer_counts(fastq_files, **jellyfish_args)

        # Delete .jf file that jellyfish creates. This file has no use outside of jellyfish
        os.remove(counts_path.replace(".counts", ".jf"))

        cgr_args = config.get("cgr", {})
        with open(counts_path, "r") as f:
            try:
                cgr.count_file_to_image_file(f, counts_path.replace(".counts", ".png"), **cgr_args)
            except Exception as e:
                if "unexpected keyword argument" in str(e):
                    bad_key = str(e).split("'")[1]
                    raise KeyError(f"CGR: '{bad_key}' is not a valid argument for generating images.")
                else:
                    raise e

        if not config.get("files", {}).get("fastq", False):
            for files in fastq_files:
                os.remove(files)

        if not config.get("files", {}).get("counts", False):
            os.remove(counts_path)



from typing import Tuple
from io import BytesIO, TextIOWrapper
import math

import numpy as np
import PIL.Image as im
import rich.progress as prog
from rich.progress import Progress


PROGBAR_COLUMNS = (
    prog.SpinnerColumn(),
    prog.TextColumn("[progress.description]{task.description}"),
    prog.BarColumn(),
    prog.TaskProgressColumn(),
    prog.TimeElapsedColumn(),
)


def contains_valid_characters(input_string: str) -> bool:
    """Checks if the given string contains only valid nucleotides characters.

    Args:
        input_string (str): The string to check for validity.

    Returns:
        bool: Returns true if and only if all characters in the input string are either 'A', 'T', 'C', or 'G'.
    """
    for char in input_string:
        if char not in {"A", "T", "C", "G"}:
            return False

    return True


def parse_count_file_line(line: str, k: int) -> Tuple[str, int]:
    """Reads a single line in a counts file and returns a parsed tuple in the form of (kmer, count)

    Args:
        line (str): The string of the line to parse
        k (int): The expected k-mer length

    Returns:
        Tuple[str, int]: A (str, int) tuple of the parsed line in the form of (kmer, count)
    """
    line_split = line.split()
    kmer = line_split[0]

    error_prefix = "CGR error reading counts file:"

    if k > 0 and len(kmer) != k:
        raise Exception(f"{error_prefix} The k-mer {kmer} does not match the reported length k={k}.")

    if not contains_valid_characters(kmer):
        raise Exception(
            f"{error_prefix} Invalid k-mer character in k-mer {kmer} (valid characters are A, T, C, G)"
        )

    try:
        count = line_split[1]

        if count == "" or count.isspace():
            raise Exception()
    except Exception:
        raise Exception(f"{error_prefix} Count not provided for k-mer: '{kmer}'")

    try:
        count = int(count)
    except ValueError:
        raise Exception(f"{error_prefix} Count for k-mer {kmer} must be an integer'")

    if count < 1:
        raise Exception(f"{error_prefix} All k-mer counts must be ≥1")

    return (kmer, count)


def calculate_pos(
    kmer: str, size: int, corner_labels: Tuple[str, str, str, str]
) -> Tuple[int, int]:
    """Returns the pixel position (x, y) of a k-mer in a CGR image with a given image size

    Args:
        kmer (str): The k-mer as a string
        size (int): The side length (in pixels) of the square image
        corner_labels (Tuple[str, str, str, str]): The labels for the corners of the image (as defined in CGR algorithm)

    Returns:
        Tuple[int, int]: A tuple representing the (x, y) pixel position of the k-mer in the image
    """

    x, y = 0, 0

    # use bit shifting instead of division to avoid floating point values
    offset = size >> 1

    bot_left, _, top_right, bot_right = corner_labels

    for base in reversed(kmer):
        if base == top_right or base == bot_right:
            x += offset

        if base == bot_left or base == bot_right:
            y += offset

        offset >>= 1

    return (x, y)


def generate_image_arr(
    counts_file: TextIOWrapper, verbose=True, size=None, log10=True, normalized=True
) -> np.ndarray:
    """Generates a numpy array representing an image covering RGB channels

    Args:
        input_data (TextIOWrapper): An open k-mer counts file
        verbose (bool, optional): Whether to print a progress bar to the console. Defaults to True.
        size (_type_, optional): The side length (in pixels) of the generated square image. Defaults to None.
        log10 (bool, optional): Whether to take the logarithm of each pixel value. Defaults to True.
        normalized (bool, optional): Whether to normalize all pixels after generating the image. Defaults to True.

    Returns:
        np.ndarray: _description_
    """

    with counts_file:
        counts_file.seek(0)

        (first_kmer, _) = parse_count_file_line(counts_file.readline(), 0)
        k = len(first_kmer)

        num_lines = sum(1 for _ in counts_file) + 1

        counts_file.seek(0)

        if size is None:
            size = 2**k

        r = np.zeros((size, size))
        g = np.zeros((size, size))
        b = np.zeros((size, size))

        with Progress(*PROGBAR_COLUMNS, disable=not verbose) as progress:
            task = progress.add_task("Generating image...\n", total=num_lines)

            for line in counts_file:
                (kmer, count) = parse_count_file_line(line, k)

                if log10:
                    count = math.log10(count)

                # weak H-bonds W = {A, T} and strong H-bonds S = {G, C} on the diagonals
                r_pos = calculate_pos(kmer, size, ("A", "G", "T", "C"))
                r[r_pos] = count

                # purine R = {A, G} and pyrimidine Y = {C, T} on the diagonals
                g_pos = calculate_pos(kmer, size, ("A", "T", "G", "C"))
                g[g_pos] = count

                # amino group M = {A, C} and keto group K = {G, T} on the diagonals
                b_pos = calculate_pos(kmer, size, ("A", "T", "C", "G"))
                b[b_pos] = count

                progress.advance(task, advance=1)

        rgb = np.dstack((r, g, b))

        if normalized:
            rgb = rgb / np.max(rgb)

        return (rgb * 255).astype(np.uint8)


def count_file_to_image(input_data: TextIOWrapper, verbose=True, **kwargs) -> im.Image:
    """
    Takes an open counts file and returns a generated image in the form of a PIL image object

    Args:
        input_data (TextIOWrapper): An open k-mer counts file
        verbose (bool, optional): Whether to print a progress bar to the console. Defaults to True.
        **kwargs: Addition keyword arguments can be provided as refernced in the generate_image_arr function.

    Returns:
        im.Image: The resulting image object
    """

    chaos = generate_image_arr(input_data, verbose, **kwargs)
    return im.fromarray(chaos)


def count_file_to_image_file(
    input_data: TextIOWrapper,
    output_file: str | BytesIO,
    output_type="png",
    verbose=True,
    **kwargs,
):
    """
    Takes an open counts file and creates an image at the provided file path or buffer.

    Args:
        input_data (TextIOWrapper): An open k-mer counts file
        output_file (str | BytesIO): The path to the image output file or the bytes buffer
        output_type (str, optional): The file type of the output image. Defaults to "png".
        verbose (bool, optional): Whether to print a progress bar to the console. Defaults to True.
        **kwargs: Addition keyword arguments can be provided as refernced in the generate_image_arr function.
    """

    img = count_file_to_image(input_data, verbose, **kwargs)
    img.save(output_file, output_type)


import os
import requests
from ftplib import FTP
from typing import List, Optional

import rich.progress as prog
from rich.progress import Progress

PROGBAR_COLUMNS = (
    prog.SpinnerColumn(),
    prog.TextColumn("[progress.description]{task.description}"),
    prog.BarColumn(),
    prog.DownloadColumn(),
    prog.TaskProgressColumn(),
    prog.TimeElapsedColumn(),
)

class NoAccessionsFoundError(Exception):
    """An error for when no accessions are able to be found in a text file or provided list."""
    pass

def get_run_accessions(accessions: List[str] | str):
    """
    Given a list of accessions (with possible project accessions) or a file path
    to a text file of accession numbers, returns a list of run accessions.

    Args:
        accessions (List[str] | str): Either a list of run or project accessions numbers 
            or a file path to a text file of accession numbers.

    Raises:
        TypeError: If the accessions is not a list of strings or a str
        FileNotFoundError: If the accessions file path does not exist.

    Returns:
        _type_: _description_
    """
    if not isinstance(accessions, (list, str)):
        raise TypeError(
            "Value for key 'accessions' in config file must be a list of strings or a string."
        )

    if isinstance(accessions, list):
        if not all(isinstance(item, str) for item in accessions):
            raise TypeError(
                "List of accessions in config file must be a list of strings."
            )

    # we know this will be a string format. we can check if it is a valid file path then parse throught the file
    elif isinstance(accessions, str):
        if accessions != "" and os.path.exists(accessions):
            # extract accession numbers from file into a list and set it to accession_list_config
            with open(accessions, "r") as f:
                acc_nums = f.readlines()
            accessions = [acc.strip() for acc in acc_nums]
        else:
            raise FileNotFoundError(
                "File path for accessions is invalid or does not exist."
            )

    run_accessions = []

    for accession in accessions:
        if accession[:3] == "PRJ":
            run_accessions.extend(get_project_accessions(accession))
        else:
            run_accessions.append(accession)

    if run_accessions == []:
        raise NoAccessionsFoundError("No accession numbers found.")

    return run_accessions


def get_project_accessions(prj_accession: str) -> List[str]:
    """
    Takes a project accession number and returns a list of run accessions associated with the project.

    Args:
        prj_accession (str): A project accession number (often starts with "PRJ").

    Returns:
        List[str]: A list of run accessions associated with the project.
    """
    url = f"https://www.ebi.ac.uk/ena/portal/api/search?result=read_run&query=study_accession={prj_accession}&fields=run_accession"
    
    try:
        response = requests.get(url)
    except ConnectionError:
        raise ConnectionError(f"Could not find provided project accessions: {prj_accession}")

    content = response.content.decode()
    lines = content.splitlines()[1:]  # ignore the header line

    # get the first value in a line (the accession)
    return [line.split("\t")[0] for line in lines]


def download_fastq(run_accession: str, output_dir: Optional[str] = None) -> List[str]:
    """
    Downloads fastq.gz files from the ENA FTP server using an accession number.
    Returns a list of the local file paths of the downloaded files.

    Args:
        run_accession (str): A valid ERR, SRR, and DRR run acccession number.
        output_dir (str, optional): The optional directory where the downloaded files will go. 
            If None, will download files to the current working directory. Defaults to None.

    Returns:
        List[str]: A list of local file paths to the downloaded files.
    """
    # small argument validations for the accession parameter
    if (
        type(run_accession) is not str
        or not run_accession.isalnum()
        or len(run_accession) < 9
        or len(run_accession) > 11
    ):
        raise ValueError(
            f"Invalid accession number: {run_accession}. \n"
            "Valid accession numbers have between 9 and 11 characters while containing only numbers or letters."
        )
    try:
        ftp = FTP("ftp.sra.ebi.ac.uk")
    except ConnectionError:
        raise ConnectionError("Failed to connect to ENA database. Please check your connection and try again.")
    
    try:
        ftp.login()

        prefix = run_accession[:6]
        suffix = run_accession[6:]

        directory = f"/vol1/fastq/{prefix}/"

        # handles different format of directory for accession numbers
        match len(suffix):
            case 3:
                directory += f"{run_accession}"
            case 4:
                directory += f"00{suffix[-1]}/{run_accession}"
            case 5:
                directory += f"0{suffix[-2:]}/{run_accession}"

        try:
            ftp.cwd(directory)
        except Exception:
            raise NotADirectoryError(
                f"Failed to access the directory for the provided accession number of {run_accession}.\n"
                "Please ensure that the accession number is correct and the corresponding FASTQ files are available on ENA."
            )

        file_names = ftp.nlst()
        if file_names == []:
            raise FileNotFoundError(f"No files found for the accession number: {run_accession}.")

        if output_dir is not None:
            if not os.path.exists(output_dir):
                raise NotADirectoryError("Output directory given for downloading files does not exist.")
            
        output_files = []

        with Progress(*PROGBAR_COLUMNS) as progress:
            for file_name in file_names:
                size = ftp.size(f"{file_name}")
                task = progress.add_task(f"Downloading {file_name}", total=size)

                # build local file path
                if output_dir is not None:
                    local_file_path = os.path.join(output_dir, file_name)
                else:
                    local_file_path = file_name

                output_files.append(local_file_path)

                # skip download if the entire file already exists
                if (
                    os.path.isfile(local_file_path)
                    and os.path.getsize(local_file_path) == size
                ):
                    progress.update(task, advance=size)
                    continue

                try:
                    with open(local_file_path, "wb") as f:

                        def callback(data):
                            f.write(data)
                            progress.update(task, advance=len(data))

                        ftp.retrbinary(f"RETR {file_name}", callback)

                except Exception:
                    raise ConnectionError(
                        f"Download failed to complete for the accession number: {run_accession}.\n"
                        "Please check your connection and try again."
                    )
        return output_files
    finally:
        ftp.close()
